import json
import sys

from discordbot.api import *
from discordbot import error

class Payload:
    def __init__(self, cereal = None, raw = None):
        if raw is None:
            try:
                self.raw = json.loads(cereal)
            except BaseException as e:
                raise error.ReceivedInvalidPayload("Invalid JSON: %s" % cereal)
        else:
            self.raw = raw
        
        if not "d" in self.raw or not "op" in self.raw:
            raise error.ReceivedInvalidPayload("Did not receive 'd' and 'op' fields: %s" % cereal)
        
        self.opcode = self.raw["op"]
        self.data = self.raw["d"]
        if opcode.get_name(self.opcode) is None:
            raise error.ReceivedInvalidPayload("Invalid opcode: %s" % self.opcode)
    
    def __str__(self):
        return json.dumps(self.raw)
    __repr__ = __str__

class Event(Payload):
    def __init__(self, pl):
        super().__init__(raw = pl.raw)
        
        if not 's' in self.raw or not 't' in self.raw:
            raise error.ReceivedInvalidPayload("Did not receive 's' and 't' fields: %s" % self)
        
        self.sequence = self.raw['s']
        self.name = self.raw['t']

class ReadyEvent(Event):
    def __init__(self, evt):
        super().__init__(evt)
        
        if not 'user' in self.data or not 'session_id' in self.data:
            raise error.ReceivedInvalidPayload("Did not receive 'user' and 'session_id' data: %s" % str(self.data))
        
        self.session_id = self.data['session_id']
        try:
            self.user = structure.User(self.data['user'])
        except error.InvalidAPIObject as e:
            raise error.ReceivedInvalidPayload("Bad user: %s" % str(e))

class GuildCreateEvent(Event):
    def __init__(self, evt):
        super().__init__(evt)
        
        try:
            self.guild = structure.Guild(self.data)
        except error.InvalidAPIObject as e:
            raise error.ReceivedInvalidPayload("Bad guild: %s" % str(e))

class MessageCreateEvent(Event):
    def __init__(self, evt):
        super().__init__(evt)
        
        try:
            self.message = structure.Message(self.data)
        except error.InvalidAPIObject as e:
            raise error.ReceivedInvalidPayload("Bad message: %s" % str(e))
            
class MessageEditEvent(Event):
    def __init__(self, evt):
        super().__init__(evt)
        
        try:
            self.before = None
            if str(self.data['id']) in structure.Message.cache:
                # make a copy of the cached version
                raw = structure.Message.cache[str(self.data['id'])].raw
                self.before = structure.Message(raw)
            
            # load new version (and cache it)
            self.after = structure.Message(self.data)
        except error.InvalidAPIObject as e:
            raise error.ReceivedInvalidPayload("Bad message: %s" % str(e))

class MessageDeleteEvent(Event):
    def __init__(self, evt):
        super().__init__(evt)
        
        try:
            self.message = None
            if str(self.data['id']) in structure.Message.cache:
                self.message = structure.Message.cache[str(self.data['id'])]
        except error.InvalidAPIObject as e:
            raise error.ReceivedInvalidPayload("Bad message: %s" % str(e))

class MessageBulkDeleteEvent(Event):
    def __init__(self, evt):
        super().__init__(evt)
        
        try:
            self.channel = structure.Channel({"id": self.data["channel_id"]})
            self.guild = structure.Guild({"id": self.data["guild_id"]})
            
            self.messages = []
            
            for id in self.data["ids"]:
                if str(id) in structure.Message.cache:
                    self.messages.append(structure.Message.cache[str(id)])
                else:
                    self.messages.append({"id": id})
        except error.InvalidAPIObject as e:
            raise error.ReceivedInvalidPayload("Bad message: %s" % str(e))

class Hello(Payload):
    def __init__(self, pl):
        super().__init__(raw = pl.raw)
        
        if not 'heartbeat_interval' in self.data:
            raise error.ReceivedInvalidPayload("Did not receive 'heartbeat_interval' field: %s" % self)
        
        self.heartbeat_interval = self.data['heartbeat_interval'] / 1000

class OutgoingPayload:
    def dumps(self):
        return json.dumps({
            'op': self.opcode,
            'd': self.data
        })

class OutgoingHeartbeat(OutgoingPayload):
    def __init__(self, sequence):
        self.opcode = opcode.heartbeat
        self.data = sequence

class OutgoingIdentify(OutgoingPayload):
    def __init__(self, token, intents, identifier):
        self.opcode = opcode.identify
        self.data = {
            "token": token,
            "properties": {
                "$os": sys.platform,
                "$browser": str(identifier),
                "$device": str(identifier),
            },
            "intents": intents
        }

class Resume(OutgoingPayload):
    def __init__(self, token, session_id, sequence):
        self.opcode = opcode.resume
        self.data = {
            "token": token,
            "session_id": session_id,
            "seq": sequence
        }


event = 0
heartbeat = 1
identify = 2
update_presence = 3
update_voice_state = 4
resume = 6
reconnect = 7
request_members = 8
session_invalidated = 9
hello = 10
heartbeat_ack = 11

_all_opcodes = {
    'event': event,
    'heartbeat': heartbeat,
    'identify': identify,
    'update_presence': update_presence,
    'update_voice_state': update_voice_state,
    'resume': resume,
    'reconnect': reconnect,
    'request_members': request_members,
    'session_invalidated': session_invalidated,
    'hello': hello,
    'heartbeat_ack': heartbeat_ack,
}
def get_name(opcode):
    if not isinstance(opcode, int):
        return None
    for name in _all_opcodes:
        if _all_opcodes[name] == opcode:
            return name
    return None


from discordbot import error
from discordbot.cache import cached
import time, json

class Snowflake:
    def __init__(self, v):
        self.valid = True
        self.v = v
        
        try:
            if not isinstance(self.v, int):
                self.v = int(self.v)
            
            self.timestamp = ((self.v >> 22) & 0x3ffffffffff + 1420070400000) / 1000.0
        except:
            self.valid = False
        
        if not self.valid:
            self.timestamp = time.time() - 3600.0
    
    def __str__(self):
        return str(self.v)

def check_requirements(obj):
    for f in obj.__class__.required_fields:
        if not f in obj.raw:
            raise error.InvalidAPIObject("'%s' object did not contain required field: %s" % (obj.__class__.__name__, f))

def set_partials(obj):
    obj.partial = False
    for f in obj.__class__.partial_fields:
        if f in obj.raw:
            obj.__dict__[f] = obj.raw[f]
        else:
            obj.__dict__[f] = None
            obj.partial = True

def merge_partials(new, old):
    partial = False
    for f in new.__class__.partial_fields:
        if new.__dict__[f] is None:
            new.__dict__[f] = old.__dict__[f]
        else:
            old.__dict__[f] = new.__dict__[f]
        
        if new.__dict__[f] is None:
            partial = True
    
    new.partial = partial
    old.partial = partial

class APIObject:
    required_fields = []
    partial_fields = []
    cache_id = "default"
    
    def cache_dump(self):
        #d = {}
        #for f in self.required_fields + self.partial_fields:
        #    d[f] = self.__dict__[f]
        #return json.dumps(d)
        return json.dumps(self.raw)
    
    def cache_load(self, d):
        #d = json.loads(d)
        #for f in d:
        #    self.__dict__[f] = d[f]
        self.__init__(json.loads(d))
    
    def update_cache(self):
        if self.cache_id in self.cache:
            merge_partials(self, self.cache[self.cache_id])
        else:
            self.cache[self.cache_id] = self

@cached
class Guild(APIObject):
    required_fields = ['id']
    partial_fields = []
    
    #def update_cache(new_object):
    #    if new_object.id in Guild.cache:
    #        old_object = Guild.cache[new_object.id]
    #        merge_partials(new_object, old_object)
    #    else:
    #        Guild.cache[new_object.id] = new_object
    
    
    def __init__(self, raw):
        self.raw = raw
        
        check_requirements(self)
        try:
            self.snowflake = Snowflake(self.raw['id'])
            self.id = str(self.snowflake)
        except BaseException as e:
            raise error.InvalidAPIObject("'%s' could not be parsed w/ error: %s" % (self.__class__.__name__, e))
        
        set_partials(self)
        #self.__class__.update_cache(self)
        self.cache_id = self.id
        self.update_cache()

@cached
class Channel(APIObject):
    required_fields = ['id']
    partial_fields = []
    
    #def update_cache(new_object):
    #    if new_object.id in Channel.cache:
    #        old_object = Channel.cache[new_object.id]
    #        merge_partials(new_object, old_object)
    #    else:
    #        Channel.cache[new_object.id] = new_object
    
    def __init__(self, raw):
        self.raw = raw
        
        check_requirements(self)
        try:
            self.snowflake = Snowflake(self.raw['id'])
            self.id = str(self.snowflake)
        except BaseException as e:
            raise error.InvalidAPIObject("'%s' could not be parsed w/ error: %s" % (self.__class__.__name__, e))
        
        set_partials(self)
        self.cache_id = self.id
        self.update_cache()
    
    def __str__(self):
        return "<#%s>" % self.id

@cached
class User(APIObject):
    required_fields = ['id']
    partial_fields = ['username', 'discriminator']
    
    #def update_cache(new_object):
    #    if new_object.id in User.cache:
    #        old_object = User.cache[new_object.id]
    #        merge_partials(new_object, old_object)
    #    else:
    #        User.cache[new_object.id] = new_object
    
    def __init__(self, raw):
        self.raw = raw
        
        check_requirements(self)
        try:
            self.snowflake = Snowflake(self.raw['id'])
            self.id = str(self.snowflake)
        except BaseException as e:
            raise error.InvalidAPIObject("'%s' could not be parsed w/ error: %s" % (self.__class__.__name__, e))
        
        set_partials(self)
        self.cache_id = self.id
        self.update_cache()
    
    def __str__(self):
        return "<@%s>" % self.id

@cached
class Message(APIObject):
    required_fields = ['id', 'author', 'channel_id', 'content']
    partial_fields = ['guild_id']
    
    #def update_cache(new_object):
    #    if new_object.id in Message.cache:
    #        old_object = Message.cache[new_object.id]
    #        merge_partials(new_object, old_object)
    #    else:
    #        Message.cache[new_object.id] = new_object
    
    def __init__(self, raw):
        self.raw = raw
        
        check_requirements(self)
        try:
            self.snowflake = Snowflake(self.raw['id'])
            self.id = str(self.snowflake)
            
            self.content = str(self.raw['content'])
        except BaseException as e:
            raise error.InvalidAPIObject("'%s' could not be parsed w/ error: %s" % (self.__class__.__name__, e))
        
        self.channel = Channel({'id': self.raw['channel_id']})
        self.author = User(self.raw['author'])
        
        set_partials(self)
        
        if self.guild_id is None:
            self.guild = None
        else:
            self.guild = Guild({'id': self.raw['guild_id']})
        
        self.cache_id = self.id
        self.update_cache()
    
    def get_url(self):
        return "https://discord.com/channels/%s/%s/%s" % (self.guild.id if not self.guild is None else "???", self.channel.id, self.id)
    
    def __str__(self):
        return "Message sent by %s in %s: %s" % (self.author, self.channel, self.content)



# default cache limits
Guild.cache.set_limit(50, 1_000) # rather small memory footprint
Channel.cache.set_limit(50, 1_000) # ditto
User.cache.set_limit(10, 1_000) # larger
Message.cache.set_limit(1_000, 100_000) # largest, but by far most important


# http
ok = 200
created = 201
no_content = 204
not_modified = 304
bad_request = 400
unauthorized = 401
forbidden = 403
not_found = 404
method_not_allowed = 405
too_many_requests = 429
gateway_unavailable = 502

http_success = [
    ok,
    created,
    no_content,
    not_modified
]

http_bad_request = [
    bad_request,
    not_found,
    method_not_allowed
]

http_insufficient_permissions = [
    unauthorized,
    forbidden
]

http_try_later = [
    too_many_requests,
    gateway_unavailable
]